body {
  background-color: #e5e7eb;
  height: 100vh;
  margin: 0;
  padding: 0;
}

body > div {
  display: flex;
  flex-direction: column;
  height: 100vh;
}

body > div > h1 {
  background-color: #1d4ed8;
  height: 40px;
  margin: 0;
  padding: 8px;
}

body > div > h1 > img {
  height: 32px;
  margin: 4px 8px 0 8px;
  width: 32px;
}

body > div > div {
  color: #888;
  font-family: monospace;
  font-size: 1.5em;
  font-weight: 500;
  padding: 8px 12px;
}

body > div > div > p {
  margin: 0 0 .25em 0;
  padding: 0;
}

body > div > div > p.loading {
  font-style: italic;
}

body > div > div > p > span.blink {
  animation: blinker 1s linear infinite;
}

body > div > div > div.small {
  font-size: .6em;
}

body > div > div > h1 {
  font-size: 1.5em;
  margin: 0 0 5px 0;
  padding: 0 0 0 10px;
}

body > div > div > h2 {
  font-size: 1em;
  margin: 0 0 10px 0;
  padding: 0 0 0 10px;
}

body > div > div > pre {
  background: #fefefe;
  border: 1px solid #ddd;
  border-radius: 8px;
  font-size: .75em;
  margin: 0 0 10px 0;
  padding: 1em .5em;
}

@keyframes blinker {
  50% {
    opacity: 0;
  }
}
