{% extends "base.html" %}
{% block title %}Initializing{% end %}
{% block content %}
<p>$ load "imbi.js", 8, 1<p>
<p class="loading">
  Initializing Imbi v{{ handler.application.settings['version'] }} <span class="blink">&hellip;</span>
</p>
{% end %}
